//Heap.java

import java.util.Vector;

/**
 * A heap is a tree-based data structure in which the data is only
 * partially sorted. 
 * <p>
 * This heap may be initialized to sort either in ascending or descending
 * order.  Each node of a heap tree contains a value >= (or <=)
 * than each of its children. Thus, the top of the tree always contains 
 * the greatest (or lowest) value.  A heap tree is always complete.
 *
 * @author Zach Tomaszewski
 * @version 13 Apr 05
 */
public class Heap {

   private Vector heap;  //keeps track of this heap's contents; 
                         //more flexible than array
   private boolean descending;  //whether this heap should sort in descending
                                //order; if false, then in ascending order
  
  /**
   * Creates a new heap, sorted in ascending order (lowest value at the top).
   */
  public Heap() {
    this(false);
  }
  
  /**
   * Creates a new heap, sorted in the specified order.
   * If set to sort in descending order, the top of heap will contain
   * the object with the greatest value in the heap.
   * 
   * @param descending  Whether to sort this tree in descending order;
   *                   otherwise sorts in ascending order.
   */
  public Heap(boolean descending) {
    this.descending = descending;
    this.heap = new Vector();
  }

  /**
   * Adds the given item into this heap.
   * @return whether the item was successfully added.
   */
  public boolean add(Comparable item) {
     this.heap.add(item);
     this.trickleUp(this.getSize() - 1);
     return true;  //will always add it
  }

  /**
   * Returns the number of nodes in this heap.
   * (This generally corresponds to the number of items in the heap.)
   */
  public int getSize() {
    return this.heap.size();
  }

  /**
   * Returns whether this heap is empty.
   */
  public boolean isEmpty() {
    return this.heap.size() == 0;    
  }

  /**
   * Returns the object at the top of this heap without removing
   * that object or otherwise affecting the heap.
   */
  public Comparable peek() throws EmptyHeapException {
    if (!isEmpty()) {
      return (Comparable) this.heap.get(0);
    }else {
      throw new EmptyHeapException();
    }
  }
  
  /**
   * Rebuilds this heap, assuming that it is in a semiheap state.
   * Semiheap state means that:
   * -- all consecutive indexes of the 
   * underlying data structure contain Comparable objects
   * (which is always true since we are currently using a Vector),
   * -- that the node in the given index position might not be greater 
   * than it's children
   * -- that the children of the given node are heaps and not semi-heaps.
   */
  private void rebuild(int index) {
    int leftChild = (index * 2) + 1;
    int rightChild = leftChild + 1;
    int greaterChild = -1;      
    int lesserChild = -1;

    /* determine whether children exist and which is greater */
    if (leftChild < this.getSize()) {
      //a left child exists at least, so for now its both greater and lesser
      greaterChild = leftChild;
      lesserChild = leftChild;
    }else { 
      //there is no real left child
      leftChild = -1;
    }
    
    if (leftChild >= 0 && rightChild < this.getSize()) {
      //there is a left child AND a right child, so which is greater?
      if (((Comparable) this.heap.get(leftChild)).compareTo(
                        this.heap.get(rightChild)) <= 0) {
        lesserChild = leftChild;
        greaterChild = rightChild;
      }else {
        lesserChild = rightChild;
        greaterChild = leftChild;
      }
    }else { //no right child
      rightChild = -1;
    }
      
    if (this.descending) {  //descending: greatest value at heap top
      //swap root with the greater child, if there is a greaterChild
      //and its value is greater than index's value
      if (greaterChild >= 0 &&  
          ((Comparable) this.heap.get(index)).compareTo(
                        this.heap.get(greaterChild)) < 0) {
        Object temp = this.heap.get(index);
        this.heap.set(index, this.heap.get(greaterChild));
        this.heap.set(greaterChild, temp);
        this.rebuild(greaterChild);
      }//otherwise no swap needed; we're done
    }else { //ascending: lowest value at top of heap
      //swap root with the lesser child, if there is a lesserChild
      //and its value is less than index's value
      if (lesserChild >= 0 &&  
          ((Comparable) this.heap.get(index)).compareTo(
                        this.heap.get(lesserChild)) > 0) {
        Object temp = this.heap.get(index);
        this.heap.set(index, this.heap.get(lesserChild));
        this.heap.set(lesserChild, temp);
        this.rebuild(lesserChild);
      }//otherwise no swap needed; we're done
    }    
  }
    
  /**
   * Removes the object at the top of this heap and returns it.
   *
   * @throws EmptyHeapException if there are no elements to remove
   */
  public Comparable remove() throws EmptyHeapException {
    if (this.getSize() == 0) {
      throw new EmptyHeapException(); 
    }else if (this.getSize() == 1) { //last element 
      return (Comparable) this.heap.remove(0);
    }else {
      Comparable obj = (Comparable) this.heap.get(0);
      //move the last element of heap to replace the top
      this.heap.set(0, this.heap.remove(this.getSize() - 1));
      this.rebuild(0);
      return obj;
    }
  }

  /**
   * Returns a one-dimensional, array-like representation of a heap.
   * Since a heap is a complete binary tree, the representation shows
   * nodes from left to right, starting at the top level and moving down
   * the tree.
   */
  public String toString() {
    return this.heap.toString();
  }

  /**
   * Compares the object at the given index with its parent.
   * If its parent is > or < (depending on the nature of this
   * heap), it will swap the two and continue to move this 
   * child up the tree.
   *
   * @param child  the index of the child node
   */
  public void trickleUp(int child) {
    if (child < this.getSize() && child > 0) { //within the heap somewhere
      int parent = (child - 1)/2;
      // IF (descending && child > parent) OR (ascending && child < parent)
      // THEN swap
      if ( (this.descending && 
           ((Comparable) this.heap.get(child)).compareTo(
                         this.heap.get(parent)) > 0) ||                         
           (!this.descending && 
           ((Comparable) this.heap.get(child)).compareTo(
                         this.heap.get(parent)) < 0) ) {
        //swap
        Object temp = this.heap.get(parent);
        this.heap.set(parent, this.heap.get(child));
        this.heap.set(child, temp);
        this.trickleUp(parent);
      }
    }
  }

  /**
   * Creates a couple sample heaps and tests them briefly,
   * printing the results to the screen.
   */
  public static void main(String[] args) {
    Heap heap = new Heap();
    heap.add(new Integer(60));
    heap.add(new Integer(8));
    heap.add(new Integer(15));
    heap.add(new Integer(35));
    heap.add(new Integer(1));
    heap.add(new Integer(6));
    heap.add(new Integer(7));
    System.out.print("Expected heap is [1 8 6 60 35 15 7]: ");
    System.out.println(heap);
    
    System.out.print("Three removes generates [1 6 7]: ");
    for (int i = 0; i < 3; i++) {
      System.out.print(heap.remove() + " ");
    }
    System.out.println();
    System.out.print("Resulting heap is [8 35 15 60]: ");
    System.out.println(heap);
    System.out.println("isEmpty? [false]: " + heap.isEmpty());
    System.out.println("size? [4]: " + heap.getSize());

    System.out.print("Four more removes generates [8 15 35 60]: ");
    for (int i = 0; i < 4; i++) {
      System.out.print(heap.remove() + " ");
    }
    System.out.println();
    System.out.println("isEmpty? [true]: " + heap.isEmpty());
    
  }

  
}//end class


