//PriorityQueue.java

/**
 * A Priority Queue ADT with Thread concurrency.
 * <p>
 * A priority queue always returns the item with the highest priority 
 * from its collection.  For this priority queue, priority can
 * be defined as "highest" or "lowest" valued item in the collection.
 * Items are compared using Comparable's compareTo method, and so
 * all objects stored in this queue must be Comparable.  
 * <p>
 * Currently implemented with a heap.
 * 
 * @author David Nickles
 * @author Zach Tomaszewski
 * @version 14 Apr 05 / v1.0
 */
class PriorityQueue implements Queue {

  private Heap heap;

  /**
   * Constructs an initially empty Priority Queue
   * with top priority defined as the element with the lowest "value".
   */
  public PriorityQueue() {
    this(true);
  }

  /**
   * Constructs an empty priority queue with the specified
   * priority definition.
   * 
   * @param lowest  If true, then the elements in this queue with the 
   *                lowest values have the highest priority; otherwise,
   *                the elements with the highest value are top priority.
   */
  public PriorityQueue(boolean lowest) {
    this.heap = new Heap(!lowest);
  }

  // ADT accessor methods

  /**
   * This method returns the count of the number of items currently residing in
   * the priority queue.
   * 
   * @return the number of items
   */
  public int size() {
    return this.heap.getSize();  
  }

  /**
   * This method returns true if this priority queue is empty,
   * or returns false if the priority queue contains any objects.
   * 
   * @return a boolean indicating the emptiness of the priority queue.
   */
  public synchronized boolean empty() {
    return this.heap.isEmpty();
  }

  /**
   * This method reveals the object of highest priority in the queue without
   * removing it.  If this queue is empty, front will wait until another thread
   * adds an item.
   * 
   * @return the highest priority object in the queue.
   */
  public synchronized Object front() {
  
    try { // wait if the priority queue is empty.
      while (size() == 0) {
        wait();
      }
    } catch (InterruptedException e) {
      //woke up
    }finally {
      return this.heap.peek();
    }
  }
  
  /**
   * Returns a String showing an "unsorted" collection of this 
   * PriorityQueue's contents.
   * (Actually, it's a String representation of this queue's internal
   * heap.)
   */
  public String toString() {
    return this.heap.toString();
  }
  
  // ADT mutator methods

  /**
   * Inserts a new item into this priority queue. This method is
   * synchronized so that only one other thread at a time can hold the lock to
   * this priority queue object.  After the item is added, will notify
   * all other threads waiting for this queue.
   * 
   * @param newItem Comparable object to be added.
   * @throws CastClassException if newItem is not Comparable
   */
  public synchronized void enqueue(Object newItem) {
    this.heap.add((Comparable) newItem);
    notifyAll(); // Let waiters know something arrived
  }

  /**
   * Removes the highest priority item from the priority queue.
   * If this queue is currently empty, dequeue will wait until 
   * an item is added to this queue by another thread.
   * 
   * @return The highest priority Comparable object.
   */
  public synchronized Object dequeue() {
  
    try { // wait if the priority queue is empty.
      while (size() == 0) {
        wait();
      }
    } catch (InterruptedException e) {
      //woke up
    }finally {
      return this.heap.remove();
    }
  }

} // end PriorityQueue class
